#!/usr/bin/python3

# Copyright (c) 2024 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import csv
import json
import logging
import os
import re
from pathlib import Path

import flatdict
import xmltodict

from osm_fieldwork.convert import Convert
from osm_fieldwork.support import basename
from osm_fieldwork.xlsforms import xlsforms_path

# Instantiate logger
log = logging.getLogger(__name__)


class ODKParsers(Convert):
    """A class to parse the CSV files from ODK Central."""

    def __init__(
        self,
        yaml: str = None,
    ):
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        xlsforms_path.replace("xlsforms", "")
        if yaml:
            pass
        else:
            pass
        self.config = super().__init__(yaml)
        self.saved = dict()
        self.defaults = dict()
        self.entries = dict()
        self.types = dict()

    def CSVparser(
        self,
        filespec: str,
        data: str = None,
    ) -> list:
        """Parse the CSV file from ODK Central and convert it to a data structure.

        Args:
            filespec (str): The file to parse.
            data (str): Or the data to parse.

        Returns:
            (list): The list of features with tags
        """
        all_tags = list()
        if not data:
            f = open(filespec, newline="")
            reader = csv.DictReader(f, delimiter=",")
        else:
            reader = csv.DictReader(data, delimiter=",")
        for row in reader:
            tags = dict()
            # log.info(f"ROW: {row}")
            for keyword, value in row.items():
                if keyword is None or value is None:
                    continue
                if len(value) == 0:
                    continue
                base = basename(keyword).lower()
                # There's many extraneous fields in the input file which we don't need.
                if base is None or base in self.ignore or value is None:
                    continue
                else:
                    # log.info(f"ITEM: {keyword} = {value}")
                    if base in self.types:
                        if self.types[base] == "select_multiple":
                            vals = self.convertMultiple(value)
                            if len(vals) > 0:
                                tags.update(vals)
                            continue
                    # When using geopoint warmup, once the display changes to the map

                    # location, there is not always a value if the accuracy is way
                    # off. In this case use the warmup value, which is where we are
                    # hopefully standing anyway.
                    if base == "latitude" and len(value) == 0:
                        if "warmup-Latitude" in row:
                            value = row["warmup-Latitude"]
                            if base == "longitude" and len(value) == 0:
                                value = row["warmup-Longitude"]
                    items = self.convertEntry(base, value)
                    # log.info(f"ROW: {base} {value}")
                    if len(items) > 0:
                        if base in self.saved:
                            if str(value) == "nan" or len(value) == 0:
                                # log.debug(f"FIXME: {base} {value}")
                                val = self.saved[base]
                                if val and len(value) == 0:
                                    log.warning(f'Using last saved value for "{base}"! Now "{val}"')
                                    value = val
                            else:
                                self.saved[base] = value
                                log.debug(f'Updating last saved value for "{base}" with "{value}"')
                        # Handle nested dict in list
                        if isinstance(items, list):
                            items = items[0]
                        for k, v in items.items():
                            tags[k] = v
                    else:
                        tags[base] = value
            # log.debug(f"\tFIXME1: {tags}")
            all_tags.append(tags)
        return all_tags

    def JSONparser(
        self,
        filespec: str = None,
        data: str = None,
    ) -> list:
        """Parse the JSON file from ODK Central and convert it to a data structure.
        The input is either a filespec to open, or the data itself.

        Args:
            filespec (str): The JSON or GeoJson input file to convert
            data (str): The data to convert

        Returns:
            (list): A list of all the features in the input file
        """
        log.debug(f"Parsing JSON file {filespec}")
        total = list()
        if not data:
            file = open(filespec, "r")
            infile = Path(filespec)
            if infile.suffix == ".geojson":
                reader = geojson.load(file)
            elif infile.suffix == ".json":
                reader = json.load(file)
            else:
                log.error("Need to specify a JSON or GeoJson file!")
                return total
        elif isinstance(data, str):
            reader = geojson.loads(data)
        elif isinstance(data, list):
            reader = data

        # JSON files from Central use value as the keyword, whereas
        # GeoJSON uses features for the same thing.
        if "value" in reader:
            data = reader["value"]
        elif "features" in reader:
            data = reader["features"]
        else:
            data = reader
        for row in data:
            # log.debug(f"ROW: {row}\n")
            tags = dict()
            if "properties" in row:
                row["properties"]  # A GeoJson formatted file
            else:
                pass  # A JOSM file from ODK Central

            # flatten all the groups into a sodk2geojson.pyingle data structure
            flattened = flatdict.FlatDict(row)
            # log.debug(f"FLAT: {flattened}\n")
            for k, v in flattened.items():
                last = k.rfind(":") + 1
                key = k[last:]
                # a JSON file from ODK Central always uses coordinates as
                # the keyword
                if key is None or key in self.ignore or v is None:
                    continue
                # log.debug(f"Processing tag {key} = {v}")
                if key == "coordinates":
                    if isinstance(v, list):
                        tags["lat"] = v[1]
                        tags["lon"] = v[0]
                        # poi = Point(float(lon), float(lat))
                        # tags["geometry"] = poi
                    continue

                if key in self.types:
                    if self.types[key] == "select_multiple":
                        # log.debug(f"Found key '{self.types[key]}'")
                        if v is None:
                            continue
                        vals = self.convertMultiple(v)
                        if len(vals) > 0:
                            tags.update(vals)
                        continue
                items = self.convertEntry(key, v)
                if items is None or len(items) == 0:
                    continue

                if type(items) == str:
                    log.debug(f"string Item {items}")
                elif type(items) == list:
                    # log.debug(f"list Item {items}")
                    tags.update(items[0])
                elif type(items) == dict:
                    # log.debug(f"dict Item {items}")
                    tags.update(items)
            # log.debug(f"TAGS: {tags}")
            if len(tags) > 0:
                total.append(tags)

        # log.debug(f"Finished parsing JSON file {filespec}")
        return total

    def XMLparser(
        self,
        filespec: str,
        data: str = None,
    ) -> list:
        """Import an ODK XML Instance file ito a data structure. The input is
        either a filespec to the Instance file copied off your phone, or
        the XML that has been read in elsewhere.

        Args:
            filespec (str): The filespec to the ODK XML Instance file
            data (str): The XML data

        Returns:
            (list): All the entries in the OSM XML Instance file
        """
        row = dict()
        if filespec:
            logging.info("Processing instance file: %s" % filespec)
            file = open(filespec, "rb")
            # Instances are small, read the whole file
            xml = file.read(os.path.getsize(filespec))
        elif data:
            xml = data
        doc = xmltodict.parse(xml)

        json.dumps(doc)
        tags = dict()
        data = doc["data"]
        flattened = flatdict.FlatDict(data)
        # total = list()
        # log.debug(f"FLAT: {flattened}")
        pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
        for key, value in flattened.items():
            if key[0] == "@" or value is None:
                continue
            # Get the last element deliminated by a dash
            # for CSV & JSON, or a colon for ODK XML.
            base = basename(key)
            log.debug(f"FLAT: {base} = {value}")
            if base in self.ignore:
                continue
            if re.search(pat, value):
                gps = value.split(" ")
                row["lat"] = gps[0]
                row["lon"] = gps[1]
                continue

            if base in self.types:
                if self.types[base] == "select_multiple":
                    # log.debug(f"Found key '{self.types[base]}'")
                    vals = self.convertMultiple(value)
                    if len(vals) > 0:
                        tags.update(vals)
                    continue
                else:
                    item = self.convertEntry(base, value)
                    if item is None or len(item) == 0:
                        continue
                    if len(tags) == 0:
                        tags = item[0]
                    else:
                        if type(item) == list:
                            # log.debug(f"list Item {item}")
                            tags.update(item[0])
                        elif type(item) == dict:
                            # log.debug(f"dict Item {item}")
                            tags.update(item)
        row.update(tags)
        return [row]



"""Index of archived XLSForm file paths."""

import os

xlsforms_path = os.path.dirname(os.path.abspath(__file__))

amenities = f"{xlsforms_path}/amenities.xls"
buildings = f"{xlsforms_path}/buildings.xls"
camping = f"{xlsforms_path}/camping.xls"
cemeteries = f"{xlsforms_path}/cemeteries.xls"
education = f"{xlsforms_path}/education.xls"
health = f"{xlsforms_path}/health.xls"
highways = f"{xlsforms_path}/highways.xls"
historical = f"{xlsforms_path}/historical.xls"
hotspring = f"{xlsforms_path}/hotspring.xls"
landusage = f"{xlsforms_path}/landusage.xls"
landuse = f"{xlsforms_path}/landuse.xls"
municipality = f"{xlsforms_path}/municipality.xls"
nature = f"{xlsforms_path}/nature.xls"
places = f"{xlsforms_path}/places.xls"
religious = f"{xlsforms_path}/religious.xls"
solidwaste = f"{xlsforms_path}/solidwaste.xls"
toilets = f"{xlsforms_path}/toilets.xls"
towns = f"{xlsforms_path}/towns.xls"
transportation = f"{xlsforms_path}/transportation.xls"
waste_collection = f"{xlsforms_path}/waste_collection.xlsx"
wastedisposal = f"{xlsforms_path}/wastedisposal.xls"
waterpoints = f"{xlsforms_path}/waterpoints.xls"
waterways = f"{xlsforms_path}/waterways.xls"



#!/usr/bin/python3

# This file has been replaced by ODKParsers(), and will be delete in the next release.

#
#   Copyright (C) 2023, 2024 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import argparse
import logging
import os
import re
import sys
from datetime import datetime
from pathlib import Path

import flatdict
import xmltodict
from geojson import Feature, FeatureCollection, dump
from shapely.geometry import Point

# Instantiate logger
log = logging.getLogger(__name__)


def main():
    """This is a program that reads in the ODK Instance file, which is in XML,
    and converts it to a GeoJson file so it can be viewed in an editor.
    """
    parser = argparse.ArgumentParser(description="Convert ODK XML instance file to GeoJson")
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-i", "--instance", required=True, help="The instance file(s) from ODK Collect")
    parser.add_argument("-o", "--outfile", default="tmp.geojson", help="The output file for JOSM")
    args = parser.parse_args()

    # if verbose, dump to the terminal
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    xmlfiles = list()
    if args.instance.find("*") >= 0:
        toplevel = Path()
        unesc = args.instance.replace("\\", "")
        for directory in toplevel.glob(unesc):
            if directory.is_dir():
                xml = os.listdir(directory)
                # There is always only one XML file per instance
                full = os.path.join(directory, xml[0])
            xmlfiles.append(full)
    else:
        toplevel = Path(args.instance)
        if toplevel.is_dir():
            # There is always only one XML file per instance
            full = os.path.join(toplevel, os.path.basename(toplevel))
            xmlfiles.append(full + ".xml")
        else:
            xmlfiles.append(args.instance)

    # print(xmlfiles)

    # These are all generated by Collect, and can be ignored
    features = list()
    for xml in xmlfiles:
        print("Processing instance file: %s" % xml)
        # odkxml = ODKInstance()
        # data = odkxml.parse(filespec=xml)
        file = open(xml, "r")
        # Instances are small, read the whole file
        data = file.read(os.path.getsize(xml))
        doc = xmltodict.parse(data)

        flattened = flatdict.FlatDict(doc["data"])
        poi = Point()
        feature = dict()
        for key, value in flattened.items():
            if key[0] == "@" or value is None:
                continue
            last = key.rfind(":") + 1
            key = key[last:]
            pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
            gps = re.findall(pat, value)
            if len(gps) > 0:
                tmp = gps[0].split(" ")
                lat = float(tmp[0])
                lon = float(tmp[1])
                poi = Point(lon, lat)
            else:
                feature[key] = value
        features.append(Feature(geometry=poi, properties=feature))
    collection = FeatureCollection(features)

    now = datetime.now()
    timestamp = f"_{now.year}_{now.month}-{now.day}-{now.hour}-{now.minute}"
    # outfile = args.instance.replace("*", "") + timestamp + ".geojson"
    # outfile = outfile.replace(" ", "")
    fout = open(args.outfile, "w")
    dump(collection, fout)

    print(f"Wrote output file {args.outfile}")


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/bin/python3

# Copyright (c) 2024 Humanitarian OpenStreetMap Team
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#
"""The async counterpart to OdkCentral.py, an ODK Central API client."""

import logging
import os
from asyncio import gather
from typing import Optional
from uuid import uuid4

import aiohttp

log = logging.getLogger(__name__)


class OdkCentral(object):
    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ):
        """A Class for accessing an ODK Central server via it's REST API.

        Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central
            session (str): Pass in an existing session for reuse.

        Returns:
            (OdkCentral): An instance of this class
        """
        if not url:
            url = os.getenv("ODK_CENTRAL_URL", default=None)
        self.url = url
        if not user:
            user = os.getenv("ODK_CENTRAL_USER", default=None)
        self.user = user
        if not passwd:
            passwd = os.getenv("ODK_CENTRAL_PASSWD", default=None)
        self.passwd = passwd
        verify = os.getenv("ODK_CENTRAL_SECURE", default=True)
        if type(verify) == str:
            self.verify = verify.lower() in ("true", "1", "t")
        else:
            self.verify = verify

        # Base URL for the REST API
        self.version = "v1"
        self.base = f"{self.url}/{self.version}/"

    def __enter__(self):
        """Sync context manager not allowed."""
        raise RuntimeError("Must be called with async context manager 'async with'")

    def __exit__(self):
        """Sync context manager not allowed."""
        raise RuntimeError("Must be called with async context manager 'async with'")

    async def __aenter__(self):
        """Async object instantiation."""
        # Header enables persistent connection, creates a cookie for this session
        self.session = aiohttp.ClientSession(
            raise_for_status=True,
            headers={"accept": "odkcentral"},
        )
        await self.authenticate()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Async object close."""
        if self.session:
            await self.session.close()

    async def authenticate(self):
        """Authenticate to an ODK Central server."""
        try:
            async with self.session.post(f"{self.base}sessions", json={"email": self.user, "password": self.passwd}) as response:
                token = (await response.json())["token"]
                self.session.headers.update({"Authorization": f"Bearer {token}"})
        except aiohttp.ClientConnectorError as request_error:
            await self.session.close()
            raise ConnectionError("Failed to connect to Central. Is the URL valid?") from request_error
        except aiohttp.ClientResponseError as response_error:
            await self.session.close()
            if response_error.status == 401:
                raise ConnectionError("ODK credentials are invalid, or may have changed. Please update them.") from response_error
            raise response_error


class OdkProject(OdkCentral):
    """Class to manipulate a project on an ODK Central server."""

    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ):
        """Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central.

        Returns:
            (OdkProject): An instance of this object
        """
        super().__init__(url, user, passwd)

    async def listForms(self, projectId: int, metadata: bool = False):
        """Fetch a list of forms in a project on an ODK Central server.

        Args:
            projectId (int): The ID of the project on ODK Central

        Returns:
            (list): The list of XForms in this project
        """
        url = f"{self.base}projects/{projectId}/forms"
        headers = {}
        if metadata:
            headers.update({"X-Extended-Metadata": "true"})
        try:
            async with self.session.get(url, ssl=self.verify, headers=headers) as response:
                self.forms = await response.json()
                return self.forms
        except aiohttp.ClientError as e:
            log.error(f"Error fetching forms: {e}")
            return []

    async def listSubmissions(self, projectId: int, xform: str, filters: dict = None):
        """Fetch a list of submission instances for a given form.

        Returns data in format:

        {
            "value":[],
            "@odata.context": "URL/v1/projects/52/forms/103.svc/$metadata#Submissions",
            "@odata.count":0
        }

        Args:
            projectId (int): The ID of the project on ODK Central
            xform (str): The XForm to get the details of from ODK Central

        Returns:
            (json): The JSON of Submissions.
        """
        url = f"{self.base}projects/{projectId}/forms/{xform}.svc/Submissions"
        try:
            async with self.session.get(url, params=filters, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Error fetching submissions: {e}")
            return {}

    async def getAllProjectSubmissions(self, projectId: int, xforms: list = None, filters: dict = None):
        """Fetch a list of submissions in a project on an ODK Central server.

        Args:
            projectId (int): The ID of the project on ODK Central
            xforms (list): The list of XForms to get the submissions of

        Returns:
            (json): All of the submissions for all of the XForm in a project
        """
        log.info(f"Getting all submissions for ODK project ({projectId}) forms ({xforms})")
        submission_data = []

        submission_tasks = [self.listSubmissions(projectId, task, filters) for task in xforms]
        submissions = await gather(*submission_tasks, return_exceptions=True)

        for submission in submissions:
            if isinstance(submission, Exception):
                log.error(f"Failed to get submissions: {submission}")
                continue
            log.debug(f"There are {len(submission['value'])} submissions")
            submission_data.extend(submission["value"])

        return submission_data


class OdkEntity(OdkCentral):
    """Class to manipulate a Entity on an ODK Central server."""

    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ) -> None:
        """Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central.

        Returns:
            (OdkEntity): An instance of this object.
        """
        super().__init__(url, user, passwd)

    async def listDatasets(
        self,
        projectId: int,
    ) -> list:
        """Get all Entity datasets (entity lists) for a project.

        JSON response:
        [
            {
                "name": "people",
                "createdAt": "2018-01-19T23:58:03.395Z",
                "projectId": 1,
                "approvalRequired": true
            }
        ]

        Args:
            projectId (int): The ID of the project on ODK Central.

        Returns:
            list: a list of JSON dataset metadata.
        """
        url = f"{self.base}projects/{projectId}/datasets/"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Error fetching datasets: {e}")
            return []

    async def listEntities(
        self,
        projectId: int,
        datasetName: str,
    ) -> list:
        """Get all Entities for a project dataset (entity list).

        JSON format:
        [
        {
            "uuid": "uuid:85cb9aff-005e-4edd-9739-dc9c1a829c44",
            "createdAt": "2018-01-19T23:58:03.395Z",
            "updatedAt": "2018-03-21T12:45:02.312Z",
            "deletedAt": "2018-03-21T12:45:02.312Z",
            "creatorId": 1,
            "currentVersion": {
                "label": "John (88)",
                "data": {
                    "field1": "value1"
                },
                "current": true,
                "createdAt": "2018-03-21T12:45:02.312Z",
                "creatorId": 1,
                "userAgent": "Enketo/3.0.4",
                "version": 1,
                "baseVersion": null,
                "conflictingProperties": null
            }
        }
        ]

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (str): The name of a dataset, specific to a project.

        Returns:
            list: a list of JSON entity metadata, for a dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Error fetching entities: {e}")
            return []

    async def getEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
    ) -> dict:
        """Get a single Entity by it's UUID for a project.

        JSON response:
        {
        "uuid": "a54400b6-49fe-4787-9ab8-7e2f56ff52bc",
        "createdAt": "2024-04-15T09:26:08.209Z",
        "creatorId": 5,
        "updatedAt": null,
        "deletedAt": null,
        "conflict": null,
        "currentVersion": {
            "createdAt": "2024-04-15T09:26:08.209Z",
            "current": true,
            "label": "test entity",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "osm_id": "1",
                "geometry": "test"
            },
            "version": 1,
            "baseVersion": null,
            "dataReceived": {
                "label": "test entity",
                "osm_id": "1",
                "geometry": "test"
            },
            "conflictingProperties": null
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (str): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity in the dataset.

        Returns:
            dict: the JSON entity details, for a specific dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}" f"/entities/{entityUuid}"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Error fetching entity: {e}")
            return {}

    async def createEntity(
        self,
        projectId: int,
        datasetName: str,
        label: str,
        data: dict,
    ) -> dict:
        """Create a new Entity in a project dataset (entity list).

        JSON request:
        {
        "uuid": "54a405a0-53ce-4748-9788-d23a30cc3afa",
        "label": "John Doe (88)",
        "data": {
            "firstName": "John",
            "age": "88"
        }
        }

        JSON response:
        {
        "uuid": "d2e03bf8-cfc9-45c6-ab23-b8bc5b7d9aba",
        "createdAt": "2024-04-12T15:22:02.148Z",
        "creatorId": 5,
        "updatedAt": None,
        "deletedAt": None,
        "conflict": None,
        "currentVersion": {
            "createdAt": "2024-04-12T15:22:02.148Z",
            "current": True,
            "label": "test entity 1",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "status": "READY",
                "geometry": "test"
            },
            "version": 1,
            "baseVersion": None,
            "dataReceived": {
                "label": "test entity 1",
                "status": "READY",
                "geometry": "test"
            },
            "conflictingProperties": None
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            label (str): Label for the Entity.
            data (dict): Key:Value pairs to insert as Entity data.

        Returns:
            dict: JSON of entity details.
                The 'uuid' field includes the unique entity identifier.
        """
        # The CSV must contain a geometry field to work
        # TODO also add this validation to uploadMedia if CSV format

        required_fields = ["geometry"]
        if not all(key in data for key in required_fields):
            msg = "'geometry' data field is mandatory"
            log.debug(msg)
            raise ValueError(msg)

        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities"
        try:
            async with self.session.post(
                url,
                ssl=self.verify,
                json={
                    "uuid": str(uuid4()),
                    "label": label,
                    "data": data,
                },
            ) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Failed to create Entity: {e}")
            return {}

    async def createEntities(
        self,
        projectId: int,
        datasetName: str,
        labelDataDict: dict,
    ) -> list:
        """Bulk create Entities in a project dataset (entity list).

        NOTE this endpoint will be redundant after Central 2024.01 release.

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            labelDataDict (dict): Mapping of Entity label:data (str:dict) to insert.

        Returns:
            list: A list of Entity detail JSONs.
                The 'uuid' field includes the unique entity identifier.
        """
        log.info(f"Bulk uploading Entities for project ({projectId}) dataset ({datasetName})")
        entity_data = []

        entity_tasks = [self.createEntity(projectId, datasetName, label, data) for label, data in labelDataDict.items()]
        log.info(f"Creating ({len(entity_tasks)}) entities for project " f"({projectId}) dataset ({datasetName})")
        entities = await gather(*entity_tasks, return_exceptions=True)

        for entity in entities:
            if not entity or isinstance(entity, Exception):
                continue
            entity_data.append(entity)

        if not entities:
            log.warning(f"No entities were uploaded for ODK project ({projectId}) dataset name ({datasetName})")

        return entity_data

    async def updateEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,
        newVersion: Optional[int] = None,
    ) -> dict:
        """Update an existing Entity in a project dataset (entity list).

        The JSON request format is the same as creating, minus the 'uuid' field.
        The PATCH will only update the specific fields specified, leaving the
            remainder.

        If no 'newVersion' param is provided, the entity will be force updated
            in place.
        If 'newVersion' is provided, this must be a single integer increment
            from the current version.

        Example response:
        {
        "uuid": "71fff014-7518-429b-b97c-1332149efe7a",
        "createdAt": "2024-04-12T14:22:37.121Z",
        "creatorId": 5,
        "updatedAt": "2024-04-12T14:22:37.544Z",
        "deletedAt": None,
        "conflict": None,
        "currentVersion": {
            "createdAt": "2024-04-12T14:22:37.544Z",
            "current": True,
            "label": "new label",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "osm_id": "1",
                "status": "new status",
                "geometry": "test",
                "project_id": "100"
            },
            "version": 3,
            "baseVersion": 2,
            "dataReceived": {
                "status": "new status",
                "project_id": "100"
            },
            "conflictingProperties": None
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity.
            label (str): Label for the Entity.
            data (dict): Key:Value pairs to insert as Entity data.
            newVersion (int): Integer version to increment to (current version + 1).

        Returns:
            dict: JSON of entity details.
                The 'uuid' field includes the unique entity identifier.
        """
        if not label and not data:
            msg = "One of either the 'label' or 'data' fields must be passed"
            log.debug(msg)
            raise ValueError(msg)

        json_data = {}
        if data:
            json_data["data"] = data
        if label:
            json_data["label"] = label

        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities/{entityUuid}"
        if newVersion:
            url = f"{url}?baseVersion={newVersion - 1}"
        else:
            url = f"{url}?force=true"

        try:
            log.info(
                f"Updating Entity ({entityUuid}) for project ({projectId}) "
                f"with params: label={label} data={data} newVersion={newVersion}"
            )
            async with self.session.patch(
                url,
                ssl=self.verify,
                json=json_data,
            ) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            log.error(f"Failed to update Entity: {e}")
            return {}

    async def deleteEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
    ) -> bool:
        """Delete an Entity in a project dataset (entity list).

        Only performs a soft deletion, so the Entity is actually archived.

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity.

        Returns:
            bool: Deletion successful or not.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities/{entityUuid}"
        log.debug(f"Deleting dataset ({datasetName}) entity UUID ({entityUuid})")
        try:
            log.info(f"Deleting Entity ({entityUuid}) for project ({projectId}) " f"and dataset ({datasetName})")
            async with self.session.delete(url, ssl=self.verify) as response:
                success = (response_msg := await response.json()).get("success", False)
                if not success:
                    log.debug(f"Server returned deletion unsuccessful: {response_msg}")
                return success
        except aiohttp.ClientError as e:
            log.error(f"Failed to delete Entity: {e}")
            return False

    async def getEntityCount(
        self,
        projectId: int,
        datasetName: str,
    ) -> int:
        """Get only the count of the Entity entries.

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.

        Returns:
            int: All entity data for a project dataset.
        """
        # NOTE returns no entity data (value: []), only the count
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}.svc/Entities?%24top=0&%24count=true"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                count = (await response.json()).get("@odata.count", None)
        except aiohttp.ClientError as e:
            log.error(f"Failed to get Entity count for project ({projectId}): {e}")
            return 0

        if count is None:
            log.debug(f"Failed to get Entity count for project ({projectId}) " f"dataset ({datasetName})")
            return 0

        return count

    async def getEntityData(
        self,
        projectId: int,
        datasetName: str,
        url_params: Optional[str] = None,
        include_metadata: Optional[bool] = False,
    ) -> dict | list:
        """Get a lightweight JSON of the entity data fields in a dataset.

        Be sure to check the latest docs to see which fields are supported for
        OData filtering:
        https://docs.getodk.org/central-api-odata-endpoints/#id3

        Example response list (include_metadata=False):
        [
            {
                "__id": "523699d0-66ec-4cfc-a76b-4617c01c6b92",
                "label": "the_label_you_defined",
                "__system": {
                    "createdAt": "2024-03-24T06:30:31.219Z",
                    "creatorId": "7",
                    "creatorName": "fmtm@hotosm.org",
                    "updates": 4,
                    "updatedAt": "2024-03-24T07:12:55.871Z",
                    "version": 5,
                    "conflict": null
                },
                "geometry": "javarosa format geometry",
                "user_defined_field2": "text",
                "user_defined_field2": "text",
                "user_defined_field3": "test"
            }
        ]

        Example response JSON where:
        - url_params="$top=205&$count=true"
        - include_metadata=True automatically due to use of $top param

        {
        "value": [
            {
            "__id": "523699d0-66ec-4cfc-a76b-4617c01c6b92",
            "label": "the_label_you_defined",
            "__system": {
                "createdAt": "2024-03-24T06:30:31.219Z",
                "creatorId": "7",
                "creatorName": "fmtm@hotosm.org",
                "updates": 4,
                "updatedAt": "2024-03-24T07:12:55.871Z",
                "version": 5,
                "conflict": null
            },
            "geometry": "javarosa format geometry",
            "user_defined_field2": "text",
            "user_defined_field2": "text",
            "user_defined_field3": "test"
            }
        ]
        "@odata.context": (
            "https://URL/v1/projects/6/datasets/buildings.svc/$metadata#Entities",
        )
        "@odata.nextLink": (
            "https://URL/v1/projects/6/datasets/buildings.svc/Entities"
            "?%24top=250&%24count=true&%24skiptoken=returnedtokenhere%3D"
        "@odata.count": 667
        }

        Info on OData URL params:
        http://docs.oasis-open.org
        /odata/odata/v4.01/odata-v4.01-part1-protocol.html#_Toc31358948

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            url_params (str): Any supported OData URL params, such as 'filter'
                or 'select'. The ? is not required.
            include_metadata (bool): Include additional metadata.
                If true, returns a dict, if false, returns a list of Entities.
                If $top is included in url_params, this is enabled by default to get
                the "@odata.nextLink" field.

        Returns:
            list | dict: All (or filtered) entity data for a project dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}.svc/Entities"
        if url_params:
            url += f"?{url_params}"
            if "$top" in url_params:
                # Force enable metadata, as required for pagination
                include_metadata = True

        try:
            async with self.session.get(url, ssl=self.verify) as response:
                response_json = await response.json()
                if not include_metadata:
                    return response_json.get("value", [])
                return response_json
        except aiohttp.ClientError as e:
            log.error(f"Failed to get Entity data: {e}")
            return {}



#!/usr/bin/python3

#
#   Copyright (C) 2020, 2021, 2022 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import argparse
import json
import logging
import os
import re
import sys

import flatdict
import xmltodict

# Instantiate logger
log = logging.getLogger(__name__)


class ODKInstance(object):
    def __init__(
        self,
        filespec: str = None,
        data: str = None,
    ):
        """This class imports a ODK Instance file, which is in XML into a
        data structure.

        Args:
            filespec (str): The filespec to the ODK XML Instance file
            data (str): The XML data

        Returns:
            (ODKInstance): An instance of this object
        """
        self.data = data
        self.filespec = filespec
        self.ignore = ["today", "start", "deviceid", "nodel", "instanceID"]
        if filespec:
            self.data = self.parse(filespec=filespec)
        elif data:
            self.data = self.parse(data)

    def parse(
        self,
        filespec: str,
        data: str = None,
    ) -> dict:
        """Import an ODK XML Instance file ito a data structure. The input is
        either a filespec to the Instance file copied off your phone, or
        the XML that has been read in elsewhere.

        Args:
            filespec (str): The filespec to the ODK XML Instance file
            data (str): The XML data

        Returns:
            (dict): All the entries in the OSM XML Instance file
        """
        row = dict()
        if filespec:
            logging.info("Processing instance file: %s" % filespec)
            file = open(filespec, "rb")
            # Instances are small, read the whole file
            xml = file.read(os.path.getsize(filespec))
        elif data:
            xml = data
        doc = xmltodict.parse(xml)

        json.dumps(doc)
        tags = dict()
        data = doc["data"]
        flattened = flatdict.FlatDict(data)
        rows = list()
        pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
        for key, value in flattened.items():
            if key[0] == "@" or value is None:
                continue
            if re.search(pat, value):
                gps = value.split(" ")
                row["lat"] = gps[0]
                row["lon"] = gps[1]
                continue

            # print(key, value)
            tmp = key.split(":")
            if tmp[len(tmp) - 1] in self.ignore:
                continue
            row[tmp[len(tmp) - 1]] = value

        return row


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-i", "--infile", required=True, help="instance data in XML format")
    args = parser.parse_args()

    os.path.basename(args.infile)

    # if verbose, dump to the terminal as well as the logfile.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    if not args.infile:
        parser.print_help()
        quit()

    inst = ODKInstance(args.infile)
    data = inst.parse(args.infile)
    # print(data)



"""Index of XLSForm file paths."""

import os

xlsforms_path = os.path.dirname(os.path.abspath(__file__))

entities_registration = f"{xlsforms_path}/entities_registration.xls"
entities_registration_xml = f"{xlsforms_path}/entities_registration.xml"

buildings = f"{xlsforms_path}/buildings.xls"



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023, 2024 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import csv
import logging
import os
import sys
from datetime import datetime

from osm_fieldwork.convert import Convert
from osm_fieldwork.support import basename
from osm_fieldwork.xlsforms import xlsforms_path

# Instantiate logger
log = logging.getLogger(__name__)


class CSVDump(Convert):
    """A class to parse the CSV files from ODK Central."""

    def __init__(
        self,
        yaml: str = None,
    ):
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        xlsforms_path.replace("xlsforms", "")
        if yaml:
            pass
        else:
            pass
        self.config = super().__init__(yaml)
        self.saved = dict()
        self.defaults = dict()
        self.entries = dict()
        self.types = dict()

    def parse(
        self,
        filespec: str,
        data: str = None,
    ) -> list:
        """Parse the CSV file from ODK Central and convert it to a data structure.

        Args:
            filespec (str): The file to parse.
            data (str): Or the data to parse.

        Returns:
            (list): The list of features with tags
        """
        all_tags = list()
        if not data:
            f = open(filespec, newline="")
            reader = csv.DictReader(f, delimiter=",")
        else:
            reader = csv.DictReader(data, delimiter=",")
        for row in reader:
            tags = dict()
            # log.info(f"ROW: {row}")
            for keyword, value in row.items():
                if keyword is None or len(value) == 0:
                    continue
                base = basename(keyword).lower()
                # There's many extraneous fields in the input file which we don't need.
                if base is None or base in self.ignore or value is None:
                    continue
                else:
                    # log.info(f"ITEM: {keyword} = {value}")
                    if base in self.types:
                        if self.types[base] == "select_multiple":
                            vals = self.convertMultiple(value)
                            if len(vals) > 0:
                                for tag in vals:
                                    tags.update(tag)
                                # print(f"BASE {tags}")
                            continue
                    # When using geopoint warmup, once the display changes to the map

                    # location, there is not always a value if the accuracy is way
                    # off. In this case use the warmup value, which is where we are
                    # hopefully standing anyway.
                    if base == "latitude" and len(value) == 0:
                        if "warmup-Latitude" in row:
                            value = row["warmup-Latitude"]
                            if base == "longitude" and len(value) == 0:
                                value = row["warmup-Longitude"]
                    items = self.convertEntry(base, value)
                    # log.info(f"ROW: {base} {value}")
                    if len(items) > 0:
                        if base in self.saved:
                            if str(value) == "nan" or len(value) == 0:
                                # log.debug(f"FIXME: {base} {value}")
                                val = self.saved[base]
                                if val and len(value) == 0:
                                    log.warning(f'Using last saved value for "{base}"! Now "{val}"')
                                    value = val
                            else:
                                self.saved[base] = value
                                log.debug(f'Updating last saved value for "{base}" with "{value}"')
                        # Handle nested dict in list
                        if isinstance(items, list):
                            items = items[0]
                        for k, v in items.items():
                            tags[k] = v
                    else:
                        tags[base] = value

                # log.debug(f"\tFIXME1: {tags}")
            all_tags.append(tags)
        return all_tags


def main():
    """Run conversion directly from the terminal."""
    parser = argparse.ArgumentParser(description="convert CSV from ODK Central to OSM XML")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument("-i", "--infile", required=True, help="The input file downloaded from ODK Central")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )
        logging.getLogger("urllib3").setLevel(logging.DEBUG)

    if args.yaml:
        csvin = CSVDump(args.yaml)
    else:
        csvin = CSVDump()

    csvin.parseXLS(args.xlsfile)
    osmoutfile = os.path.basename(args.infile.replace(".csv", ".osm"))
    csvin.createOSM(osmoutfile)

    jsonoutfile = os.path.basename(args.infile.replace(".csv", ".geojson"))
    csvin.createGeoJson(jsonoutfile)

    log.debug("Parsing csv files %r" % args.infile)
    data = csvin.parse(args.infile)
    # This OSM XML file only has OSM appropriate tags and values
    nodeid = -1000
    for entry in data:
        feature = csvin.createEntry(entry)
        if len(feature) == 0:
            continue
        if "refs" in feature:
            refs = list()
            for ref in feature["refs"]:
                now = datetime.now().strftime("%Y-%m-%dT%TZ")
                if len(ref) == 0:
                    continue
                coords = ref.split(" ")
                print(coords)
                node = {"attrs": {"id": nodeid, "version": 1, "timestamp": now, "lat": coords[0], "lon": coords[1]}, "tags": dict()}
                csvin.writeOSM(node)
                refs.append(nodeid)
                nodeid -= 1

            feature["refs"] = refs
            csvin.writeOSM(feature)
        else:
            # Sometimes bad entries, usually from debugging XForm design, sneak in
            if "lat" not in feature["attrs"]:
                log.warning("Bad record! %r" % feature)
                continue
            csvin.writeOSM(feature)
            # This GeoJson file has all the data values
            csvin.writeGeoJson(feature)
            # print("TAGS: %r" % feature['tags'])

    csvin.finishOSM()
    csvin.finishGeoJson()
    log.info("Wrote OSM XML file: %r" % osmoutfile)
    log.info("Wrote GeoJson file: %r" % jsonoutfile)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

#
# Copyright (C) 2020, 2021, 2022, 2023, 2024 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import argparse
import glob
import logging
import os
import sys
from pathlib import Path

from osm_fieldwork.parsers import ODKParsers
from osm_fieldwork.support import OutSupport

# Instantiate logger
log = logging.getLogger(__name__)


def main():
    """This is a program that reads in the ODK Instance file, which is in XML,
    and converts it to an OSM XML file so it can be viewed in an editor.
    """
    parser = argparse.ArgumentParser(description="Convert ODK XML instance file to OSM XML format")
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument("-i", "--infile", required=True, help="The input file")
    # parser.add_argument("-o","--outfile", default='tmp.csv', help='The output file for JOSM')
    args = parser.parse_args()

    # if verbose, dump to the terminal
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    toplevel = Path(args.infile)
    odk = ODKParsers(args.yaml)
    odk.parseXLS(args.xlsfile)
    out = OutSupport()
    xmlfiles = list()
    data = list()
    # It's a wildcard, used for XML instance files
    if args.infile.find("*") >= 0:
        log.debug(f"Parsing multiple ODK XML files {args.infile}")
        toplevel = Path(args.infile[:-1])
        for dirs in glob.glob(args.infile):
            xml = os.listdir(dirs)
            full = os.path.join(dirs, xml[0])
            xmlfiles.append(full)
        for infile in xmlfiles:
            tmp = odk.XMLparser(infile)
            entry = odk.createEntry(tmp[0])
            data.append(entry)
    elif toplevel.suffix == ".xml":
        # It's an instance file from ODK Collect
        log.debug(f"Parsing ODK XML files {args.infile}")
        # There is always only one XML file per infile
        full = os.path.join(toplevel, os.path.basename(toplevel))
        xmlfiles.append(full + ".xml")
        tmp = odk.XMLparser(args.infile)
        # odki = ODKInstance(filespec=args.infile, yaml=args.yaml)
        entry = odk.createEntry(tmp)
        data.append(entry)
    elif toplevel.suffix == ".csv":
        log.debug(f"Parsing csv files {args.infile}")
        for entry in odk.CSVparser(args.infile):
            data.append(odk.createEntry(entry))
    elif toplevel.suffix == ".json":
        log.debug(f"Parsing json files {args.infile}")
        for entry in odk.JSONparser(args.infile):
            data.append(odk.createEntry(entry))

    # Write the data
    out.WriteData(toplevel.stem, data)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2023, 2024 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import json
import logging

# import pandas as pd
import sys
from pathlib import Path

import flatdict
import geojson

from osm_fieldwork.convert import Convert

log = logging.getLogger(__name__)


class JsonDump(Convert):
    """A class to parse the JSON files from ODK Central or odk2geojson."""

    def __init__(
        self,
        yaml: str = None,
    ):
        """A class to convert the JSON file from ODK Central, or the GeoJson
        file created by the odk2geojson utility.

        Args:
            yaml (str): The filespec of the YAML config file

        Returns:
            (JsonDump): An instance of this object
        """
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        self.config = super().__init__(yaml)

    def parse(
        self,
        filespec: str = None,
        data: str = None,
    ) -> list:
        """Parse the JSON file from ODK Central and convert it to a data structure.
        The input is either a filespec to open, or the data itself.

        Args:
            filespec (str): The JSON or GeoJson input file to convert
            data (str): The data to convert

        Returns:
            (list): A list of all the features in the input file
        """
        log.debug(f"Parsing JSON file {filespec}")
        total = list()
        if not data:
            file = open(filespec, "r")
            infile = Path(filespec)
            if infile.suffix == ".geojson":
                reader = geojson.load(file)
            elif infile.suffix == ".json":
                reader = json.load(file)
            else:
                log.error("Need to specify a JSON or GeoJson file!")
                return total
        elif isinstance(data, str):
            reader = geojson.loads(data)
        elif isinstance(data, list):
            reader = data

        # JSON files from Central use value as the keyword, whereas
        # GeoJSON uses features for the same thing.
        if "value" in reader:
            data = reader["value"]
        elif "features" in reader:
            data = reader["features"]
        else:
            data = reader
        for row in data:
            # log.debug(f"ROW: {row}\n")
            tags = dict()
            # Extract the location regardless of what the tag is
            # called.
            # pat = re.compile("[-0-9.]*, [0-9.-]*, [0-9.]*")
            # gps = re.findall(pat, str(row))
            # tmp = list()
            # if len(gps) == 0:
            #     log.error(f"No location data in: {row}")
            #     continue
            # elif len(gps) == 1:
            #     # Only the warmup has any coordinates.
            #     tmp = gps[0].split(" ")
            # elif len(gps) == 2:
            #     # both the warmup and the coordinates have values
            #     tmp = gps[1].split(" ")

            # if len(tmp) > 0:
            #     lat = float(tmp[0][:-1])
            #     lon = float(tmp[1][:-1])
            #     geom = Point([lon, lat])
            #     row["geometry"] = geom
            #     # tags["geometry"] = row["geometry"]

            if "properties" in row:
                row["properties"]  # A GeoJson formatted file
            else:
                pass  # A JOSM file from ODK Central

            # flatten all the groups into a sodk2geojson.pyingle data structure
            flattened = flatdict.FlatDict(row)
            for k, v in flattened.items():
                last = k.rfind(":") + 1
                key = k[last:]
                # a JSON file from ODK Central always uses coordinates as
                # the keyword
                if key is None or key in self.ignore or v is None:
                    continue
                log.debug(f"Processing tag {key} = {v}")
                if key == "coordinates":
                    if isinstance(v, list):
                        tags["lat"] = v[1]
                        tags["lon"] = v[0]
                        # poi = Point(float(lon), float(lat))
                        # tags["geometry"] = poi
                    continue

                if key in self.types:
                    if self.types[key] == "select_multiple":
                        # log.debug(f"Found key '{self.types[key]}'")
                        if v is None:
                            continue
                        vals = self.convertMultiple(v)
                        if len(vals) > 0:
                            for tag in vals:
                                tags.update(tag)
                            # print(f"BASE {tags}")
                        continue

                items = self.convertEntry(key, v)
                if items is None or len(items) == 0:
                    continue

                if type(items) == str:
                    log.debug(f"string Item {items}")
                else:
                    log.debug(f"dict Item {items}")
                    if len(items) == 0:
                        tags.update(items[0])
            # log.debug(f"TAGS: {tags}")
            if len(tags) > 0:
                total.append(tags)

        # log.debug(f"Finished parsing JSON file {filespec}")
        return total


def main():
    """Run conversion directly from the terminal."""
    parser = argparse.ArgumentParser(description="convert JSON from ODK Central to OSM XML")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument("-i", "--infile", required=True, help="The input file downloaded from ODK Central")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )
        logging.getLogger("urllib3").setLevel(logging.DEBUG)

    if args.yaml:
        jsonvin = JsonDump(args.yaml)
    else:
        jsonin = JsonDump()

    jsonin.parseXLS(args.xlsfile)

    base = Path(args.infile).stem
    osmoutfile = f"{base}.osm"
    jsonin.createOSM(osmoutfile)

    jsonoutfile = f"{base}.geojson"
    jsonin.createGeoJson(jsonoutfile)

    log.debug("Parsing json files %r" % args.infile)
    data = jsonin.parse(args.infile)
    # This OSM XML file only has OSM appropriate tags and values
    nodeid = -1000
    for entry in data:
        feature = jsonin.createEntry(entry)
        if len(feature) == 0:
            continue
        if "refs" in feature:
            refs = list()
            for ref in feature["refs"]:
                now = datetime.now().strftime("%Y-%m-%dT%TZ")
                if len(ref) == 0:
                    continue
                coords = ref.split(" ")
                print(coords)
                node = {"attrs": {"id": nodeid, "version": 1, "timestamp": now, "lat": coords[0], "lon": coords[1]}, "tags": dict()}
                jsonin.writeOSM(node)
                refs.append(nodeid)
                nodeid -= 1

            feature["refs"] = refs
            jsonin.writeOSM(feature)
        else:
            # Sometimes bad entries, usually from debugging XForm design, sneak in
            if "lat" not in feature["attrs"]:
                log.warning("Bad record! %r" % feature)
                continue
            jsonin.writeOSM(feature)
            # This GeoJson file has all the data values
            jsonin.writeGeoJson(feature)
            # print("TAGS: %r" % feature['tags'])

    jsonin.finishOSM()
    jsonin.finishGeoJson()
    log.info("Wrote OSM XML file: %r" % osmoutfile)
    log.info("Wrote GeoJson file: %r" % jsonoutfile)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

#
#   Copyright (C) 2020, 2021, 2022 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import argparse
import logging
import os
import sys

# Instantiate logger
log = logging.getLogger(__name__)


class ODKForm(object):
    """Support for parsing an XLS Form, currently a work in progress..."""

    def __init__(self):
        """Returns:
        (ODKForm): An instance of this object.
        """
        self.fields = dict()
        self.nodesets = dict()
        self.groups = dict()
        self.ignore = ("label", "@appearance", "hint", "upload")

    def parseSelect(
        self,
        select: dict,
    ):
        """Parse a select statement in XML.

        Args:
            select (dict): The select in XML:

        Returns:
            (dict): The data from the select
        """
        print("parseSelect %r" % type(select))
        newsel = dict()
        if "item" in select:
            data = self.parseItems(select["item"])
            ref = os.path.basename(select["@ref"])
            for key in data:
                if key in self.ignore:
                    continue
                newsel[ref] = data
            print("\tQQQQQ %r" % (newsel))
        return newsel

    def parseItems(
        self,
        items: list,
    ):
        """Parse the items in a select list.

        Args:
            items (list): The select items list in XML:

        Returns:
            (dict): The data from the list of items
        """
        print("\tparseItems: %r: %r" % (type(items), items))
        newitems = list()
        # if type(items) == OrderedDict:
        #     data = list()
        #     data.append(items)
        # else:
        #     data = items

        for values in items:
            newitems.append(values["value"])

            # if type(values) == str:
            #     continue

        #     val = values['label']['@ref'].replace("/data/", "")
        #     tmp = val.split('/')
        #     group = tmp[0].replace("jr:itext(\'", "")
        #     fields = len(tmp)
        #     if fields > 2:
        #         subgroup = tmp[1]
        #         label = tmp[2].replace(":label\')", "")
        #     else:
        #         subgroup = None
        #         label = tmp[1].replace(":label\')", "")
        #     # print("VALUES: %r / %r / %r" % (group, subgroup, label))
        #     if subgroup not in newdata:
        #         newdata[subgroup] = list()
        #     #newdata[subgroup].append(label)
        #     newitems.append(label)
        # return group, subgroup, newitems
        return newitems

    def parseGroup(
        self,
        group: dict(),
    ):
        """Convert the XML of a group into a data structure.

        Args:
            group (dict): The group data
        """
        print("\tparseGroup %r" % (type(group)))
        if type(group) == list:
            for _val in group:
                for k in group:
                    print("\nZZZZ1 %r" % (k))
        else:  # it's a list
            for keyword, data in group.items():
                # FIXME: for now,. ignore media files
                if keyword in self.ignore:
                    continue
                print("WWW3 %r, %r" % (keyword, type(data)))
                # pat = re.compile('select[0-9]*')
                # if pat.match(keyword):
                if keyword[0:6] == "select":
                    print("WWW4 select")
                    self.parseSelect(data)
                    # self.groups[] =


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    parser = argparse.ArgumentParser(description="convert CSV from ODK Central to OSM XML")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-i", "--infile", help="The input file downloaded from ODK Central")
    args = parser.parse_args()

    # if verbose, dump to the terminal as well as the logfile.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    odkform = ODKForm()
    odkform.parse(args.infile)

    print("---------------------------")
    odkform.dump()
    print("---------------------------")
